@@ -35,6 +35,8 @@
 #include "gpdf-g-switch.h"
 #include "GPOutputDev.h"
 #include "PDFDoc.h"
+#include "Link.h"
+#include "gpdf-links-canvas-layer.h"
 
 BEGIN_EXTERN_C
 
@@ -43,6 +45,8 @@ struct _GPdfViewPrivate {
 	GPOutputDev *output_dev;
 	PDFDoc *pdf_doc;
 
+	GPdfLinksCanvasLayer *links_layer;
+
 	GnomeCanvasItem *page_background;
 	GnomeCanvasItem *page_shadow;
 
@@ -66,6 +70,8 @@ GPDF_CLASS_BOILERPLATE (GPdfView, gpdf_view, GnomeCanvas, PARENT_TYPE);
 enum {
 	ZOOM_CHANGED_SIGNAL,
 	PAGE_CHANGED_SIGNAL,
+	CLOSE_REQUESTED_SIGNAL,
+	QUIT_REQUESTED_SIGNAL,
 	LAST_SIGNAL
 };
 
@@ -107,37 +113,39 @@ gpdf_view_page_background_resize (GPdfView *gpdf_view,
 }
 
 static void
-setup_page_transform (gdouble *transform, gdouble width, gdouble height, gint angle)
+setup_page_transform (gdouble *transform, 
+		      gdouble x1, gdouble y1, gdouble x2, gdouble y2,
+		      gint angle)
 {
+	double width;
+	double height;
+	double rotate [6];
+	double translate [6];
+
+	width = x2 - x1;
+	height = y2 - y1;
+
+	art_affine_identity (transform);
+	art_affine_flip (transform, transform, FALSE, TRUE);
+	art_affine_rotate (rotate, angle);
+	art_affine_multiply (transform, transform, rotate);
+
 	if (angle == 90) {
-		transform [0] = 0.0;
-		transform [1] = 1.0;
-		transform [2] = 1.0;
-		transform [3] = 0.0;
 		transform [4] = 0.0;
 		transform [5] = 0.0;
 	} else if (angle == 180) {
-		transform [0] = -1.0;
-		transform [1] = 0.0;
-		transform [2] = 0.0;
-		transform [3] = 1.0;
 		transform [4] = width;
 		transform [5] = 0.0;
 	} else if (angle == 270) {
-		transform [0] = 0.0;
-		transform [1] = -1.0;
-		transform [2] = -1.0;
-		transform [3] = 0.0;
 		transform [4] = height;
 		transform [5] = width;
 	} else /* if (angle == 0) */ {
-		transform [0] = 1.0;
-		transform [1] = 0.0;
-		transform [2] = 0.0;
-		transform [3] = -1.0;
 		transform [4] = 0.0;
 		transform [5] = height;
 	}
+
+	art_affine_translate (translate, -x1, -y1);
+	art_affine_multiply (transform, translate, transform);
 }	
 
 static gint
@@ -153,17 +161,116 @@ canonical_multiple_of_90 (gint n)
 	return 90 * (gint)((n / 90.0) + .5);
 }
 
+static void
+gpdf_view_link_action_goto (GPdfView *gpdf_view, LinkGoTo *link_goto)
+{
+	PDFDoc *pdf_doc;
+	LinkDest *dest;
+	GString *named_dest;
+
+	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
+	g_return_if_fail (dynamic_cast <LinkGoTo *> (link_goto) != NULL);
+
+	pdf_doc = gpdf_view->priv->pdf_doc;
+	dest = link_goto->getDest ();
+	named_dest = link_goto->getNamedDest ();
+
+	if (dest != NULL) {
+		dest = dest->copy();
+	} else if (named_dest != NULL) {
+		named_dest = named_dest->copy();
+		dest = pdf_doc->findDest (named_dest);
+		delete named_dest;
+	}
+
+	if (dest != NULL) {
+		Ref page_ref;
+		int page;
+
+		if (dest->isPageRef ()) {
+			page_ref = dest->getPageRef ();
+			page = pdf_doc->findPage (page_ref.num, page_ref.gen);
+		} else {
+			page = dest->getPageNum ();
+		}
+
+		delete dest;
+
+		if (page != gpdf_view->priv->requested_page)
+			gpdf_view_goto_page (gpdf_view, page);
+	}	
+}
+
+static void
+gpdf_view_link_action_named (GPdfView *gpdf_view, LinkNamed *link_named)
+{
+	const char *action_name;
+
+	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
+	g_return_if_fail (dynamic_cast <LinkNamed *> (link_named) != NULL);
+
+	action_name = link_named->getName ()->getCString ();
+	
+	if (strcmp (action_name, "NextPage") == 0) {
+		gpdf_view_page_next (gpdf_view);
+	} else if (strcmp (action_name, "PrevPage") == 0) {
+		gpdf_view_page_prev (gpdf_view);
+	} else if (strcmp (action_name, "FirstPage") == 0) {
+		gpdf_view_page_first (gpdf_view);
+	} else if (strcmp (action_name, "LastPage") == 0) {
+		gpdf_view_page_last (gpdf_view);
+	} else if (strcmp (action_name, "Close") == 0) {
+		g_signal_emit (
+			G_OBJECT (gpdf_view), CLOSE_REQUESTED_SIGNAL, 0);
+	} else if (strcmp (action_name, "Quit") == 0) {
+		g_signal_emit (G_OBJECT (gpdf_view), QUIT_REQUESTED_SIGNAL, 0);
+	} else {
+		g_message ("Unimplemented named link action: %s", action_name);
+	}
+}
+
+static void
+link_clicked_cb (GPdfLinksCanvasLayer *links_layer, Link *link,
+		 gpointer user_data)
+{
+	GPdfView *gpdf_view;
+	LinkAction *link_action;
+
+	g_return_if_fail (GPDF_IS_VIEW (user_data));
+	g_return_if_fail (dynamic_cast <Link *> (link) != NULL);
+
+	gpdf_view = GPDF_VIEW (user_data);
+	link_action = link->getAction ();
+
+	switch (link_action->getKind ()) {
+	case actionGoTo:
+		gpdf_view_link_action_goto (
+			gpdf_view, dynamic_cast <LinkGoTo *> (link_action));
+		break;
+	case actionNamed:
+		gpdf_view_link_action_named (
+			gpdf_view, dynamic_cast <LinkNamed *> (link_action));
+		break;
+	default:
+		g_message ("Click on link: unimplemented Action");
+		break;
+	}
+}
+
 static gboolean
 gpdf_view_render_page (GPdfView *gpdf_view)
 {
 	GPdfViewPrivate *priv;
 	GnomePrintConfig *config;
 	PDFDoc *pdf_doc;
+	Page *pdf_page;
+	const PDFRectangle *rect;
 	gint page;
 	gint page_rotate;
 	gdouble page_width_rot, page_height_rot;
 	ArtDRect region;
 	gdouble page_transform [6];
+	Object obj;
 
 	g_return_val_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view), FALSE);
 
@@ -180,12 +287,13 @@ gpdf_view_render_page (GPdfView *gpdf_view)
 	}
 
 	pdf_doc = priv->pdf_doc;
+	pdf_page = pdf_doc->getCatalog ()->getPage (page);
+	rect = pdf_page->getBox ();
 
 	page_rotate = canonical_multiple_of_90 (pdf_doc->getPageRotate (page));
 
 	setup_page_transform (page_transform,
-			      pdf_doc->getPageWidth (page),
-			      pdf_doc->getPageHeight (page),
+			      rect->x1, rect->y1, rect->x2, rect->y2,
 			      page_rotate);
 
 	if (page_rotate == 90 || page_rotate == 270) {
@@ -218,6 +326,30 @@ gpdf_view_render_page (GPdfView *gpdf_view)
 
 	priv->current_page = page;
 
+	if (priv->links_layer) {
+		g_object_unref (G_OBJECT (priv->links_layer));
+		gtk_object_destroy (GTK_OBJECT (priv->links_layer));
+		priv->links_layer = NULL;
+	}
+
+	priv->links_layer = GPDF_LINKS_CANVAS_LAYER (gnome_canvas_item_new (
+		gnome_canvas_root (GNOME_CANVAS (gpdf_view)),
+		GPDF_TYPE_LINKS_CANVAS_LAYER,
+		"links",
+		new Links (pdf_page->getAnnots (&obj), 
+			   pdf_doc->getCatalog ()->getBaseURI ()),
+		NULL));
+	g_object_ref (G_OBJECT (priv->links_layer));
+	gtk_object_sink (GTK_OBJECT (priv->links_layer));
+
+	obj.free ();
+
+	gnome_canvas_item_affine_relative (
+		GNOME_CANVAS_ITEM (priv->links_layer), page_transform);
+
+	g_signal_connect (G_OBJECT (priv->links_layer), "link_clicked",
+			  G_CALLBACK (link_clicked_cb), gpdf_view);
+
 	return FALSE;
 }
 
@@ -528,6 +660,12 @@ gpdf_view_button_press_event (GtkWidget *widget, GdkEventButton *event)
 
 	priv = GPDF_VIEW (widget)->priv;
 
+	if (GNOME_CALL_PARENT_WITH_DEFAULT (GTK_WIDGET_CLASS,
+					    button_press_event,
+					    (widget, event), FALSE) == TRUE) {
+		return TRUE;
+	}
+
 	if (!GTK_WIDGET_HAS_FOCUS (widget))
 		gtk_widget_grab_focus (widget);
 
@@ -557,10 +695,8 @@ gpdf_view_button_press_event (GtkWidget *widget, GdkEventButton *event)
 		gdk_cursor_unref (cursor);
 		return TRUE;
 	default:
-		break;
+		return FALSE;
 	}
-
-	return FALSE;
 }
 
 static gboolean
@@ -581,22 +717,20 @@ gpdf_view_button_release_event (GtkWidget *widget, GdkEventButton *event)
 	return TRUE;
 }
 
-static gboolean
-gpdf_view_motion_notify_event (GtkWidget *widget, GdkEventMotion *event)
+static void
+gpdf_view_handle_drag_motion (GPdfView *gpdf_view, GdkEventMotion *event)
 {
 	GPdfViewPrivate *priv;
 	gint x, y, dx, dy;
 	GdkModifierType mod;
 
-	g_return_val_if_fail (GPDF_IS_NON_NULL_VIEW (widget), FALSE);
-	
-	priv = GPDF_VIEW (widget)->priv;
+	g_return_if_fail (GPDF_IS_NON_NULL_VIEW (gpdf_view));
 
-	if (!priv->dragging)
-		return FALSE;
+	priv = gpdf_view->priv;
 
 	if (event->is_hint) {
-		gdk_window_get_pointer (widget->window, &x, &y, &mod);
+		gdk_window_get_pointer (GTK_WIDGET (gpdf_view)->window,
+					&x, &y, &mod);
 	} else {
 		x = (gint)event->x;
 		y = (gint)event->y;
@@ -605,17 +739,28 @@ gpdf_view_motion_notify_event (GtkWidget *widget, GdkEventMotion *event)
 	dx = priv->drag_anchor_x - x;
 	dy = priv->drag_anchor_y - y;
 
-	gnome_canvas_scroll_to (GNOME_CANVAS (widget), 
+	gnome_canvas_scroll_to (GNOME_CANVAS (gpdf_view), 
 				priv->drag_offset_x + dx,
 				priv->drag_offset_y + dy);
 
 	priv->drag_anchor_x = (gint)event->x;
 	priv->drag_anchor_y = (gint)event->y;
-	gnome_canvas_get_scroll_offsets (GNOME_CANVAS (widget),
+	gnome_canvas_get_scroll_offsets (GNOME_CANVAS (gpdf_view),
 					 &priv->drag_offset_x,
 					 &priv->drag_offset_y);
+}
 
-	return TRUE;
+static gboolean
+gpdf_view_motion_notify_event (GtkWidget *widget, GdkEventMotion *event)
+{
+
+	g_return_val_if_fail (GPDF_IS_NON_NULL_VIEW (widget), FALSE);
+	
+	if (GPDF_VIEW (widget)->priv->dragging)
+		gpdf_view_handle_drag_motion (GPDF_VIEW (widget), event);
+
+	return GNOME_CALL_PARENT_WITH_DEFAULT (
+		GTK_WIDGET_CLASS, motion_notify_event, (widget, event), TRUE);
 }
 
 static void
@@ -702,6 +847,11 @@ gpdf_view_dispose (GObject *object)
 		priv->print_context = NULL;
 	}
 
+	if (priv->links_layer) {
+		g_object_unref (priv->links_layer);
+		priv->links_layer = NULL;
+	}
+
 	if (priv->gesture_handler) {
 		gtk_gesture_handler_destroy (priv->gesture_handler);
 		priv->gesture_handler = NULL;
@@ -758,6 +908,24 @@ gpdf_view_class_init (GPdfViewClass *klass)
 		NULL, NULL,
 		gpdf_marshal_VOID__INT,
 		G_TYPE_NONE, 1, G_TYPE_INT);
+
+	gpdf_view_signals [CLOSE_REQUESTED_SIGNAL] = g_signal_new (
+		"close_requested",
+		G_TYPE_FROM_CLASS (object_class),
+		G_SIGNAL_RUN_LAST,
+		G_STRUCT_OFFSET (GPdfViewClass, close_requested),
+		NULL, NULL,
+		gpdf_marshal_VOID__VOID,
+		G_TYPE_NONE, 0);
+
+	gpdf_view_signals [QUIT_REQUESTED_SIGNAL] = g_signal_new (
+		"quit_requested",
+		G_TYPE_FROM_CLASS (object_class),
+		G_SIGNAL_RUN_LAST,
+		G_STRUCT_OFFSET (GPdfViewClass, quit_requested),
+		NULL, NULL,
+		gpdf_marshal_VOID__VOID,
+		G_TYPE_NONE, 0);
 }
 
 static void
